-- 文字列加工

-- 文字数を取得（バイト数ではない）
local function GetStringCount(self, str)
    return mbstrlen(str or '')
--[[
    -- 頑張って作ったのに標準で用意されとるやんけ
    local result = ''
    local strLength = 0
    for c in string.gmatch(str, "[%z\1-\127\194-\244][\128-\191]*") do
        result = result..c
        strLength = strLength+1
    end

    return strLength
--]]
end

-- 文字数を指定してカット
local function CutString(...)
    local self, str, len, text = ...
    text = text or '...'
    local result = ''
    local strLength = 0
    for c in string.gmatch(str, "[%z\1-\127\194-\244][\128-\191]*") do
        if strLength >= len then
            result = result..text
            break
        end
        result = result..c
        strLength = strLength+1
    end
    -- 最後の1文字だけ省略になる場合は元の文字を返す
    if GetStringCount(self, str) == strLength+1 then
        result = str
    end
    return result
end

-- バイト数を指定してカット
local function CutByte(...)
    local self, str, len, text = ...
    text = text or '...'
    local strlen = string.len(str)
    if len >= strlen then
        return str
    end
    local result = string.sub(str, 1, len)
    for i=len+1, strlen do
        local c = string.sub(str, i, i)
        if not string.match(c, "[\128-\191]") then
            result = result..((strlen > string.len(result)) and text or '')
            break
        end
        result = result..c
    end
    return result
end

-- 文字数を指定するが極力幅を近くカット
local function CutWidth(...)
    local self, str, len, text = ...
    local cut = CutString(self, str, len)
    local cutLen = string.len(cut)
    local size = math.max(len - (cutLen-len)/3 + 1, len/2)
    return CutString(self, str, size, text)
end

-- 3桁区切り表示にする（max 999,999,999）
local function NumberFormat(self, value)
    -- return commify(string.format('%07d', value), ',', '.')
    -- commify() だけを使うと先頭のゼロ埋めがされない
    local ret = ''
    if value >= 1000000 then
        ret = ret .. math.floor(value / 1000000) .. ','
    else
        ret = '0,'
    end
    if value >= 1000 then
        ret = ret .. string.format('%03d', math.floor((value%1000000) / 1000)) .. ','
    else
        ret = ret .. '000,'
    end
    ret = ret .. string.format('%03d', value%1000)
    return ret
end

-- 1文字目のみ大文字にする
local function UpperFirst(self, text)
    return string.upper(string.sub(text, 1, 1))..string.lower(string.sub(text, 2))
end

-- ScreenXXXXOptionsからXXXX部分を取り出し、キャメルケースの大文字部分前にスペースを入れる
local function OptionName(self, text, ...)
    local prefix = ... or 'Options'
    -- まずScreenとOptions、Serviceの文字を消す
    local displayName = string.gsub(string.gsub(string.gsub(text, 'Screen', ''), 'Options', ''), 'Service', '')
    -- 1文字目を除く大文字の前にスペースを入れる
    displayName = string.gsub(string.gsub(displayName, '([A-Z][^A-Z]+)', ' %1'), '^ (.+)', '%1')
    if prefix and prefix ~= '' then
        -- 最後にPrefixの文字を付けて大文字化
        displayName = displayName..((displayName ~= '') and ' ' or '')..'Options'
    end
    return string.upper(displayName)
end

-- ダンスポイント用のフォーマットにする（100%の時だけ小数点無し）
local function DPFormat(self, ...)
    local value, plus = ...
    local prefix = ((plus and value >= 0) and '+' or '')
    if math.abs(value) >= 1 then
        return prefix..math.floor(value*100)..'%'
    end
    return prefix..FormatPercentScore(value)
end

-- ファイルを読み込む
local function ReadFile(self, path)
    if not FILEMAN:DoesFileExist(path) then
        return ''
    end
    local f = RageFileUtil.CreateRageFile()
    f:Open(path, 1)
    local ret = f:Read()
    f:Close()
    f:destroy()
    return ret
end

-- ファイルに保存
local function WriteFile(self, path, data)
    local f = RageFileUtil.CreateRageFile()
    f:Open(path, 2)
    local ret = f:Write(data)
    f:Close()
    f:destroy()
    W:Flush(path)
    return
end

-- 2つの数値の差分とp1>p2なら青、p1<p2なら赤、ほぼ同じなら灰色を返却
-- p1：基準
-- p2：対象
-- return：差分、色
local function GetPlusMinusAndColor(self, my, target, ...)
    local isNoPlay = ...
    local diffPercent = my - target
    local lead = W_TARGET:DiffToLead(diffPercent)
    local col  = (lead <= -1)
                and BoostColor(Color('Blue'), 1.33)
                or ((lead >= 1) 
                    and color('#ff4c4cff')
                    or color('#a0a0a0ff'))
    if isNoPlay then col = color('#a0a0a0ff') end
    return diffPercent, col
end

-- バイナリを16進数に直したMD5を返却
local function StringMD5(self, text)
    local md5bin = CRYPTMAN:MD5String(text)
    local ret = ''
    for i=1, string.len(md5bin) do
        local b = string.byte(md5bin, i, i)
        ret = string.format('%s%x', ret, b)
    end
    return ret
end

return {
    GetStringCount = GetStringCount,
    CutString      = CutString,
    CutByte        = CutByte,
    CutWidth       = CutWidth,
    Number         = NumberFormat,
    UpperFirst     = UpperFirst,
    OptionName     = OptionName,
    DP             = DPFormat,
    ReadFile       = ReadFile,
    WriteFile      = WriteFile,
    DiffAndColor   = GetPlusMinusAndColor,
    StringMD5      = StringMD5,
}

--[[
String.lua

Copyright (c) 2021 A.C

This software is released under the MIT License.
https://opensource.org/licenses/mit-license.php
--]]
